(* ::Package:: *)

(*
   BuTools_MAP Package
*)


BeginPackage["BuTools`MAP`"];
RAPToMAP::usage = "RAPToMAP [ matrix0, matrix1 ] -> [ outputmx0, outputmx1 ] : Similarity transforms the (matrix0, matrix1) representation to the Markovian representation (outputmx0, outputmx1), if possible."
MarginalMomentsFromMAP::usage = "MarginalMomentsFromMAP [ matrix0, matrix1, k ] -> [ moments ] : Calculates the first k marginal moments of the MAP with representaion (matrix0, matrix1)"
MarginalMomentsFromRAP::usage = "MarginalMomentsFromMAP [ matrix0, matrix1, k ] -> [ moments ] : Calculates the first k marginal moments of the RAP with representaion (matrix0, matrix1)"
LagkJointMomentsFromRAP::usage = "LagkJointMomentsFromRAP [ matrix0, matrix1, k ] -> [ moments ] : Calculates the matrix of the E(X_0^i,X_k^j) moments of the RAP with representaion (matrix0, matrix1)"
LagCorrelationsFromRAP::usage = "LagCorrelationsFromRAP [ matrix0, matrix1 ] -> [ lagcorrelations ] : Calculates the lag correlations of the RAP with representaion (matrix0, matrix1)"
RAPFromMoments::usage = "RAPFromMoments [ moments, nn ] -> [ outputmx0, outputmx1 ] : Calculates a RAP representation based on 2n-1 moments and the n*n matrix of the lag1 joint moments"
RAPFromMomentsAndCorrelations::usage = "RAPFromMomentsAndCorrelations [ moments, lagkcorrelations ] -> [ outputmx0, outputmx1 ] : Calculates a RAP representation based on 2n-1 moments and 2n-3 lag-k correlations using the K. Mitchell, A. van de Liefvoort method"
MAP2Canonical::usage = "MAP2Canonical [ matrix0, matrix1 ] -> [ matrix0, matrix1 ]  : Calculates the canonical representation of the input MAP of size 2 using the similarity tranfromation which sets the 0 elements of the canonical representation."
MAP2CanonicalExplicit::usage = "MAP2CanonicalExplicit [ matrix0, matrix1 ] -> [ matrix0, matrix1 ]  : Calculates the canonical representation of the input MAP of size 2 using the explicite expressions from the Canonical MAP2 paper"
StairCase::usage = "StairCase [ matrix0, matrix1, closing vector ] -> [ size, simularity matrix ] : Computes a smaller representation of RAP using staircase algorithm"
RAPObsMinimize::usage = "RAPObsMinimize [ matrix0, matrix1 ] -> [ matrix0, matrix1 ] : Computes a minimal representation of an MRAP using the staircase methor by eliminating the redundancy caused by the initial vector."
RAPContMinimize::usage = "RAPContMinimize [ matrix0, matrix1 ] -> [ matrix0, matrix1 ] : Computes a minimal representation of an MRAP using the staircase methor by eliminating the redundancy caused by the closing vector."
RAPMinimize::usage = "RAPMinimize [ matrix0, matrix1 ] -> [ matrix0, matrix1 ] : Computes a minimal representation of an MRAP using the staircase method by eliminating the redundancy caused by the initial vector and by the closing vector. It executes RAPContMinimize and RAPObsMinimize one after the other."



Unprotect[RAPToMAP, MarginalMomentsFromMAP, MarginalMomentsFromRAP,
LagkJointMomentsFromRAP, LagCorrelationsFromRAP, RAPFromMoments, RAPFromMomentsAndCorrelations, MAP2Canonical,
StairCase, MAP2CanonicalExplicit, RAPObsMinimize, RAPContMinimize, RAPMinimize];


Begin["`Private`"]

Needs["BuTools`Utilities`"]
Needs["BuTools`PH`"]



RAPToMAP[h0_,h1_]:= 
Module[{size,\[Epsilon]},
size=Dimensions[h0][[1]];
\[Epsilon]=10^-8;  (* numerical precision *)
If[ Max[Abs[(h0+h1).Table[1,{size}] ]]> \[Epsilon],
Print["Rowsum is different from 0 (precision:", \[Epsilon]//N, ") !!"];
Return[{h0,h1}]
];
X={h0,h1};
Return[RationalToMarkovian[ X,size, MAPTransFunc, MAPEvalFunc, MAPCheckFunc]];
];



MAPCheckFunc[X_] :=CheckMAP[X[[1]],X[[2]]];
MAPTransFunc[X_,i_,j_,a_] :=
Module[ {bb,size},
size=Dimensions[X[[1]]][[1]];
bb=IdentityMatrix[size];
bb[[i,j]]=a;
bb[[i,i]]=1-a;
Return[{Inverse[bb].X[[1]].bb,Inverse[bb].X[[2]].bb}]
];
MAPEvalFunc[X_,Error_] :=
Module[ {i,j,sum,size, h0,h1},
size=Dimensions[X[[1]]][[1]];
h0=X[[1]];
h1=X[[2]];
sum=0;
For[i=1,i<=size,i++,For[j=1,j<=size,j++,
If[i!=j,sum+=Error[h0[[i,j]]  ] ];];];
For[i=1,i<=size,i++,
  For[j=1,j<=size,j++,
    sum+=Error[ h1[[i,j]] ];
  ];
];
Return[sum];
];


MarginalMomentsFromMAP[d0_,d1_,k_]:=
Module[{\[Alpha],d0i},
If[Not[CheckMAP[d0,d1]], Print["Marginal moments of improper MAP!!"]];
MarginalMomentsFromRAP[d0,d1,k];
]



MarginalMomentsFromRAP[d0_,d1_,k_]:=
Module[{\[Alpha],d0i},
If[Not[CheckRAP[d0,d1]], Print["Marginal moments of improper RAP!!"]];
d0i=Inverse[-d0];
\[Alpha]=DRPSolve[d0i.d1] ;
Return[MomentsFromME[\[Alpha],d0, k]];
]


LagkJointMomentsFromRAP[d0_,d1_,k_]:=
Module[{h,n,\[Pi],d0i,res,i,j,Pk},
If[Not[CheckRAP[d0,d1]], Return[{{0}}]];
n = Dimensions[d1][[1]];
h=Table[1,{n}];
d0i=Inverse[-d0];
\[Pi]=DRPSolve[d0i.d1] ;
Pk=MatrixPower[d0i.d1,k];
res=Table[  (i-1)!(j-1)!\[Pi].MatrixPower[d0i,i-1].
Pk.MatrixPower[d0i,j-1].h,{i,n},{j,n}];
Return[res];
]



LagCorrelationsFromRAP[d0_,d1_]:=
Module[{e,n,\[Pi],d0i,\[Lambda],res,i},
n = Dimensions[d1][[1]];
e=Table[1,{n}];
\[Pi]=CRPSolve[d0+d1] ;
d0i=Inverse[-d0];
\[Lambda]=\[Pi]. d1 . e;
res={};
For[i=1,i<=2n-3,i=i+1,

res=Append[res,(\[Lambda]  (\[Pi].MatrixPower[d0i.d1,i].d0i.e)-1)/(2 \[Lambda] (\[Pi].d0i.e)-1)]
];
Return[res];
]



(*calculates the Appie RAP representation based on 2n-1 moments and the nxn matrix of the lag1 joint moments *) 
RAPFromMoments[mom_,nn_]:=
Module[{size,hh,gammav,gamma1,v,h0,h0i,i,j},

{v,h0}=MEFromMoments[mom];
h0i=Inverse[-h0];
size=Dimensions[v][[1]];
hh=Table[1,{size}];

gammav=Table[0,{size},{size}];
For[i=1,i<=size,i++,gammav[[i]]=(i-1)!  v.MatrixPower[h0i,i-1];];
gamma1=Table[0,{size},{size}];
For[i=1,i<=size,i++,gamma1[[i]]=(i-1)!  MatrixPower[h0i,i-1].hh];
gamma1=Transpose[gamma1];

Return[{h0,-h0.Inverse[gammav].nn.Inverse[gamma1]}];
] 




(*calculates a RAP representation based on 2n-1 moments and 2n-3 autocorrelations
using the K.Mitchell,A.van de Liefvoort method *) 
RAPFromMomentsAndCorrelations[mom_,corr_]:=
Module[{size,kk,rmom,rcorr,yy,ii,rsize, hh,e1,tt,uu,i,j,tti,uui,d0,d1},
rmom=Table[mom[[i]]/i!,{i,Dimensions[mom][[1]]}];
{kk,size}=Appie[rmom,False] ;

If[Dimensions[corr]<2 size -3 ,
Print["there are less then 2n-3 correlation parameters"]; Return[{{0},{{0}}}]
];

rcorr=Part[corr,1;;2 size -3]/((rmom[[2]]-rmom[[1]]^2)/(2 rmom[[2]]-rmom[[1]]^2));
{yy,rsize}=Appie[rcorr,False] ;
If[(rsize +1)!= size ,
Print["correlation order is different from ME order"]; Return[{{0},{{0}}}]
];
ii=IdentityMatrix[ size];
ii[[2;; size,2;; size]]=yy;
yy=ii;

hh=Table[1,{size},{1}];
e1=Table[0,{1},{size}];
e1[[1,1]]=1;

tt=Table[0,{size},{size}];
For[i=1,i<=size,i++,For[j=1,j<=i,j++,tt[[i,j]]= 1;];];
tti=Inverse[tt];
(*Print[tt//MatrixForm,tti//MatrixForm];*)

uu=Table[0,{size},{size}];
For[i=1,i<=size,i++,For[j=i,j<=size,j++,uu[[i,j]]= 1/(size-i+1);];];
uui=Inverse[uu];
(*Print[uu//MatrixForm,uui//MatrixForm];*)

d0=-Inverse[uui.tt.kk.tti.uu];
(* d1=- d0.uui.tt.yy.tti.uu; *)
d1=- d0.uui.tt.yy.tti.uu;
Return[{d0,d1}];
] 


MAP2Canonical[d0_,d1_] :=
Module[ {\[Gamma],W,w1,w2, Wi,res},
If[Not[Dimensions[d0][[1]] ==2],If[BuToolsVerbose==True,Print["MAP2Canonical: size is not 2!"]]; Return[{{{0}},{{0}}}]
];
 If[Not[CheckMAP[d0,d1]], Return[{{{0}},{{0}}}]]; 

\[Gamma]=Eigenvalues[Inverse[-d0//N].d1];
(* Print["eigenvalues of P: ",\[Gamma]]; *)

W=({
 {1-w1, w1},
 {w2, 1-w2}
});
Wi=Inverse[W];

If[\[Gamma][[2]]>= 0,
 (* Print["positive correlation!!!"]; *)
res=Solve[{(Wi.d0.W)[[2,1]]==0,(Wi.d1.W)[[1,2]]==0}, {w1,w2}];
If[CheckMAP[(Wi.d0.W)/.res[[1]],(Wi.d1.W)/.res[[1]]],
 Return[{(Wi.d0.W)/.res[[1]],(Wi.d1.W)/.res[[1]]}],
Return[{(Wi.d0.W)/.res[[2]],(Wi.d1.W)/.res[[2]]}]
];
];

If[ \[Gamma][[2]]<0,
 (* Print["negative correlation!!!"]; *)
res=Solve[{(Wi.d0.W)[[2,1]]==0,(Wi.d1.W)[[1,1]]==0}, {w1,w2}];
If[CheckMAP[(Wi.d0.W)/.res[[1]],(Wi.d1.W)/.res[[1]]],
 Return[{(Wi.d0.W)/.res[[1]],(Wi.d1.W)/.res[[1]]}],
Return[{(Wi.d0.W)/.res[[2]],(Wi.d1.W)/.res[[2]]}]
];
];

];



StairCase[Xin_,Yin_,Zin_]:=
Module[{\[Epsilon],X,Y,Z,m,U,ranksum,crit,r,Ui,S,T,TEMP,Z1,X1,X2,X3,X4,Y1,Y2,Y3,Y4,Transf,I,n,x,i,nonzero,zeros,R,
y,\[CapitalGamma],TEMP1,sizeUi},
\[Epsilon]=10^-8;  (* numerical precision *)
Print["Numerical precision of StairCase is:", \[Epsilon]//N, " !!" ]; 
X=Xin;Y=Yin;Z=Zin//N;
m=Dimensions[X][[1]];
U=IdentityMatrix[m];
ranksum=0;
crit=True;
While[crit, 
r=MatrixRank[Z];
ranksum+=r;
{Ui,S,T}=SingularValueDecomposition[Z];
TEMP=Transpose[Ui].Z;
Z1=TEMP[[1;;r,All]];

TEMP=Transpose[Ui].X.Ui;
X1=TEMP[[1;;r,1;;r]];
X2=TEMP[[1;;r,r+1;;-1]];
X3=TEMP[[r+1;;-1,1;;r]];
X4=TEMP[[r+1;;-1,r+1;;-1]];

TEMP=Transpose[Ui].Y.Ui;
Y1=TEMP[[1;;r,1;;r]];
Y2=TEMP[[1;;r,r+1;;-1]];
Y3=TEMP[[r+1;;-1,1;;r]];
Y4=TEMP[[r+1;;-1,r+1;;-1]];

sizeUi=Dimensions[Ui][[1]];
Transf=IdentityMatrix[ranksum-r+sizeUi];
Transf[[-sizeUi ;;-1,-sizeUi ;;-1]]=Transpose[Ui];
U=Transpose[Transf.Transpose[U]];


X=X4;Y=Y4;
Z=Join[X3,Y3,2];

(* Print["Z=",Z]; *)
If[Max[SingularValueList[Z]]<\[Epsilon] ||MatrixRank[Z]==m-ranksum,crit=False];
];

I=IdentityMatrix[m];

If[Max[SingularValueList[Z]]<\[Epsilon],
n=ranksum;
x=Transpose[U].Table[  1,{m},{1}];
x=x[[1;;n,All]];

R=IdentityMatrix[n];
If[Min[Abs[x]]<\[Epsilon],
nonzero=0;
zeros=Table[False,{n}];
For[i=1,i<=n,i++,
If[Abs[x[[i,1]]]<\[Epsilon],zeros[[i]]=True,nonzero=i];
];
(* Print["nonzero=",nonzero, "; zeros=",zeros ]; *)

If[nonzero !=0,
For[i=1,i<=n,i++, If[zeros[[i]],R[[i,nonzero]]=1]
]; 
];
];

y=R.x;
(* Print["x=",x, "; R=",R,"; y=",y]; *)

\[CapitalGamma]=DiagonalMatrix[Flatten[y]];
TEMP=I;
TEMP[[1;;n,1;;n]]=Inverse[\[CapitalGamma]];
TEMP1=I;
TEMP1[[1;;n,1;;n]]=R;

Return[{n,Inverse[TEMP.TEMP1.Transpose[U]]}];
];

If[MatrixRank[Z]==m-ranksum,Return[{m,I}]];
]



MAP2CanonicalExplicit[d0_,d1_] :=
Module[ {\[Epsilon],\[Mu]1,\[Mu]2,\[Mu]3,h1,h2,h3,\[Gamma], g0,g1,\[Lambda]1,\[Lambda]2,\[Alpha],p,a,b,W,w1,w2, Wi,res},
\[Epsilon]=10^-8;
If[Not[Dimensions[d0][[1]] ==2],If[BuToolsVerbose==True,Print["MAP2Canonical: size is not 2!"]]; Return[{{{0}},{{0}}}]
];
 If[Not[CheckMAP[d0,d1]], Return[{{{0}},{{0}}}]]; 

{\[Mu]1,\[Mu]2,\[Mu]3}=MarginalMomentsFromRAP[d0,d1,3];
h1=\[Mu]1;h2=\[Mu]2/(2 \[Mu]1^2)-1;h3=\[Mu]3/(6 \[Mu]1^3)-\[Mu]2^2/(4 \[Mu]1^4);
If[Abs[h2] < \[Epsilon],If[BuToolsVerbose==True,Print["MAP2Canonical: degenerate case: exponential marginal!"]]; Return[{{{0}},{{0}}}]
];


If[h2> 0,
\[Lambda]1=(h3+h2^2+h2-Sqrt[(h3+h2^2+h2)^2-4 h2 h3])/(2 h1 h3);
\[Lambda]2=(h3+h2^2+h2+Sqrt[(h3+h2^2+h2)^2-4 h2 h3])/(2 h1 h3);
p=(-h3-h2^2+h2+Sqrt[(h3+h2^2+h2)^2-4 h2 h3])/(h3+h2^2+h2+Sqrt[(h3+h2^2+h2)^2-4 h2 h3]); \[Alpha]=\[Lambda]1/\[Lambda]2;
];

If[h2< 0,
\[Lambda]1=(h3+h2^2+h2+Sqrt[(h3+h2^2+h2)^2-4 h2 h3])/(2 h1 h3);
\[Lambda]2=(h3+h2^2+h2-Sqrt[(h3+h2^2+h2)^2-4 h2 h3])/(2 h1 h3);
p=(h3+h2^2-h2+Sqrt[(h3+h2^2+h2)^2-4 h2 h3])/(-h3-h2^2-h2+Sqrt[(h3+h2^2+h2)^2-4 h2 h3]); \[Alpha]=\[Lambda]1/\[Lambda]2;
];

\[Gamma]=Eigenvalues[Inverse[-d0//N].d1];
(* Print["eigenvalues of P: ",\[Gamma]]; *)
\[Gamma]=\[Gamma][[2]];

If[\[Gamma]>= 0,
 (* Print["positive correlation!!!"]; *)
a=1/(2 \[Alpha]) (1+\[Alpha] \[Gamma] -p (1-\[Gamma])-Sqrt[(1+\[Alpha] \[Gamma] -p (1-\[Gamma]))^2-4 \[Alpha] \[Gamma]]);
b=1/2  (1+\[Alpha] \[Gamma] -p (1-\[Gamma])+Sqrt[(1+\[Alpha] \[Gamma] -p (1-\[Gamma]))^2-4 \[Alpha] \[Gamma]]);

g0=({
 {-\[Lambda]1, (1-a)\[Lambda]1},
 {0, -\[Lambda]2}
});
g1=({
 {a \[Lambda]1, 0},
 {(1-b) \[Lambda]2, b \[Lambda]2}
});
];

If[ \[Gamma]<0,
 (* Print["negative correlation!!!"]; *)
a=-\[Gamma]/(p (1-\[Gamma])- \[Alpha] \[Gamma]);
b=p (1-\[Gamma])- \[Alpha] \[Gamma];

g0=({
 {-\[Lambda]1, (1-a)\[Lambda]1},
 {0, -\[Lambda]2}
});
g1=({
 {0, a \[Lambda]1},
 {b \[Lambda]2, (1-b) \[Lambda]2}
});
];
Return[{g0,g1}];
];



RAPMinimize[d0in_,d1in_]:= 
Module[ {d0,d1,size, vec,B ,n},
d0=d0in;d1=d1in;
{d0,d1}=RAPContMinimize[d0,d1]//Chop ;
{d0,d1}=RAPObsMinimize[d0,d1] //Chop ;
Return[{d0,d1}];
]



RAPContMinimize[d0in_,d1in_]:= 
Module[ {d0,d1,size, vec,B ,n},
d0=d0in;d1=d1in;
size=Dimensions[d0][[1]];
vec= Table[1,{size},{1}];
{n,B}=StairCase[d0,d1,vec];
d0=Inverse[B].d0.B;
d1=Inverse[B].d1.B;
d0=d0[[1;;n,1;;n]];
d1=d1[[1;;n,1;;n]];
Return[{d0,d1}];
]


RAPObsMinimize[d0in_,d1in_]:= 
Module[ {d0,d1,size, vec,B ,n},
d0=d0in;d1=d1in;
size=Dimensions[d0][[1]];
vec={DRPSolve[Inverse[-d0].d1]};
{n,B}=StairCase[Transpose[d0],Transpose[d1],Transpose[vec]];
d0=Inverse[B].d0.B;
d1=Inverse[B].d1.B;
d0=d0[[1;;n,1;;n]];
d1=d1[[1;;n,1;;n]];
Return[{d0,d1}];
]


End[(* Private *)]

Protect[RAPToMAP, MarginalMomentsFromMAP, MarginalMomentsFromRAP, 
LagkJointMomentsFromRAP, LagCorrelationsFromRAP, RAPFromMoments, RAPFromMomentsAndCorrelations, MAP2Canonical,
StairCase, MAP2CanonicalExplicit, RAPObsMinimize, RAPContMinimize, RAPMinimize];

EndPackage[]
